/* LaTeX2Sdoc
 *  Copyright (C) 2000-2001 SAKURAI, Masashi 
 * (m.sakurai@cmt.phys.kyushu-u.ac.jp)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package jp.gr.java_conf.ccs2.tool.tex2sdoc;
import java.util.*;

/**
   Manage TeX macro system.
 */
public class MacroManager {

    Hashtable table = new Hashtable();
    Vector macros = new Vector();
    Macro [] array = null;

    public MacroManager() {
    }

    public void addMacro(Macro m) {
	table.put(m.getName(),m);
	macros.addElement(m);
	array = null;
    }

    static String DEF = "\\def";
    static String NEWCOMMAND = "\\newcommand";
    
    /**
       make a macro object from tex command and add it to this.
       Ex, \def\cmdname{ ~ }
           \newcommand{\cmdname}[n]{ ~~ }
     */
    public void addMacro(String def) {
	int ps = def.indexOf( DEF );
	if (ps != -1) {
	    //DEF command
	    ps += DEF.length();
	    int br = TeXUtil.searchBrace(def,'{',ps);
	    if (br == -1) {
		throw new RuntimeException(
		    "Wrong definition of new tex command. ["+def+"]");
	    }
	    String name = def.substring(ps,br);
	    ArgumentInfo info = TeXUtil.parseMatchedBrace(def,br,1);
	    if (info == null) {
		throw new RuntimeException(
		    "Wrong definition of new tex command. ["+def+"]");
	    }
	    String content = info.contents[0];
	    addMacro(new Macro(name,content));
	    return;
	}

	ps = def.indexOf( NEWCOMMAND );
	if (ps != -1) {
	    //NEWCOMMAND command
	    ps += NEWCOMMAND.length();
	    int argnumber = 0;
	    ArgumentInfo cmd = TeXUtil.parseMatchedBrace(def,ps,2);
	    if (cmd == null) {
		throw new RuntimeException(
		    "Wrong definition of new tex command. ["+def+"]");
	    }
	    String name = cmd.contents[0];
	    String content = cmd.contents[1];
	    ArgumentInfo argn = TeXUtil.parseMatchedBrace(def,'[',']',ps,1);
	    if (argn != null) {
		try {
		    argnumber = Integer.parseInt(argn.contents[0]);
		}catch(NumberFormatException e) {}
	    }
	    addMacro(new Macro(name,argnumber,content));
	    return;
	}
	System.err.println("Wrong definition of new tex command. ["+def+"]");
    }
    

    public Macro getMacro(String m) {
	return (Macro)table.get(m);
    }

    public Macro [] getMacros() {
	if (array != null) return array;
	array = new Macro [macros.size()];
	for (int i=0;i<macros.size();i++) {
	    array[i] = (Macro)macros.elementAt(i);
	}
	return array;
    }

    /** expand macro from first position */
    public String expandMacro(String input) {
	return expandMacro(input,0);
    }

    /** expand macro from given position */
    public String expandMacro(String input,int sx) {
	return shortProcessMacro(input,sx);
    }

    /**
       sample code
     */
    public static void main(String [] args) {
	MacroManager mm = new MacroManager();
	mm.addMacro(new Macro("\\dd","\\mathrm{d}\\."));
	mm.addMacro(new Macro("\\dvec",2,"\\ddd{#1}_{#2}"));
	mm.addMacro(new Macro("\\ddd",1,"\\dd{\\mathbf{#1}}"));
	mm.addMacro("\\def\\Hline{\\noalign{\\hrule height1pt}}");
	mm.addMacro("\\newcommand{\\cfrac}[2]{\\frac{\\displaystyle #1}{\\displaystyle #2}}");
	
	String a = "\\int a x + \\cfrac{\\dd y}{\\dd x} \\dd x";
	String b = "\\int r^2 \\sin(\\theta) g(r) \\dvec{r}{1}";
	String c = "\\hline \\Hline";

	System.out.println(mm.expandMacro(a));
	System.out.println(mm.expandMacro(b));
	System.out.println(mm.expandMacro(c));
    }


    /** (short input-text version)
       @param text input tex source
       @param ps start position to process
       @return expanded text
     */
    String shortProcessMacro(String text, int ps) {
	Macro [] ms = getMacros();
	if (ms == null || ms.length == 0) return text;
	while(true) {
	    int sx = -1;
	    Macro macro = null;
	    for (int i=0;i<ms.length;i++) {
		Macro m = ms[i];
		int xps = TeXUtil.findCmd(text,m.getName(),ps);
		if (xps == -1) continue;
		if (xps < sx || sx == -1) {
		    sx = xps;
		    macro = m;
		}
	    }
	    if (sx == -1) break;
	    //
	    text = processMacro(
		text,sx,macro,TeXUtil.parseMatchedBrace(
		    text,sx,macro.getArgNumber()));
	}
	return text;
    }

    //=== ( tex parse utility ) =======


    /** 
	@param text input tex source.
	@param ps   start position of finded macro.
	(point to the first position, the char "/" )
	@paran m    macro object
	@param args arguments of the macro (accept null)
	@return expanded output 
     */
    static String processMacro(String text,int ps,
				   Macro m,ArgumentInfo args) {
	StringBuffer ret = new StringBuffer(text.substring(0,ps));
	if ( args == null ) {
	    ret.append(m.process(null));
	    ret.append(text.substring(ps+m.getName().length()));
	} else {
	    ret.append(m.process(args.contents));
	    ret.append(text.substring(args.endPos));
	}
	
	return ret.toString();
    }

}

