/* LaTeX2Sdoc
 *  Copyright (C) 2000-2001 SAKURAI, Masashi (m.sakurai@cmt.phys.kyushu-u.ac.jp)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package jp.gr.java_conf.ccs2.tool.tex2sdoc.math;
import java.util.*;
import jp.gr.java_conf.ccs2.tool.tex2sdoc.TeXUtil;
import jp.gr.java_conf.ccs2.util.StringUtil;

/**
   This object treats a foundamental function of translation from 
   methematical presentation of TeX to MathML. 
   Because TeX has so many presentation commands,
   I prepare the plugin framework for any TeX commands.
   @see jp.gr.java_conf.ccs2.tool.tex2sdoc.math.MathMLCommandHandler
 */
public class MathMLVisitor implements TeXSourceVisitor, OutputState {

    //for public
    public MathMLVisitor() {}

    //for copy
    protected MathMLVisitor(MathMLVisitor parent) {
	setOneline( parent.isOneline() );
	setEntity( parent.isEntity() );
	setNameSpace( parent.getNameSpace() );
	setMathMLURI( parent.getMathMLURI() );
	setStartTab( parent.getCurrentTab()+1 );
    }

    protected String mathmlUri = "http://www.w3.org/1998/Math/MathML";
    protected String nameSpace = null;

    protected StringBuffer buffer;//output buffer
    protected int startTab = 1;//left tab limit 
    protected int curTab = 0;//current tab depth

    protected Stack endStack = new Stack();//string stack in end-brace
    protected Stack handlerStack = new Stack();//string stack in end of command
    protected Stack interruptStack = new Stack();//string stack in end of any tag

    //status switch
    // if true, this visitor object is in <mtext> content.
    protected boolean textState = false;

    protected String log = "";

    //==(switch & option)========

    public void setMathMLURI(String s) { mathmlUri = s; }
    public String getMathMLURI() {return mathmlUri;}
    public String getXMLURI() {return getMathMLURI();}

    /** set name space, such as "mathml:" */
    public void setNameSpace(String n) {
	if (n == null) return;
	if (n.indexOf(":")!=-1)
	    nameSpace = n;
	else
	    nameSpace = n+":";
    }
    public String getNameSpace() {return nameSpace;}

    protected boolean entity = false;
    public void setEntity(boolean b) {entity = b;}
    public boolean isEntity() {return entity; }

    boolean oneline = false;
    public void setOneline(boolean b) {oneline = b;  }
    public boolean isOneline() {return oneline; }

    //==(code)========

    /** get result string */
    public String getCode() {
	curTab = 0;
	return 
	    TagUtil.makeBtagWithURI(this,"math",getMathMLURI())+
	    buffer.toString()+
	    TagUtil.makeEtag(this,"math");
    }

    protected String getInsiderCode() {
	return buffer.toString();
    }

    /** get translation error log */
    public String getLog() {return log;}

    //===(visitor stuff)===

    public void init() {
	endStack.removeAllElements();
	handlerStack.removeAllElements();
	curTab = 1;
	log = "";
	buffer = new StringBuffer();
    }

    public void identifier(String id) throws ParseException {
	if (isTextState()) {
	    textOutput(id);
	    return;
	}
	//split into one character
	String [] mm = new String [id.length()];
	for (int i=0;i<id.length();i++) {
	    identifier_gen(id.substring(i,i+1));
	}
    }

    void identifier_gen(String id) {
	addString( TagUtil.makeTag(this,"mi",id));
    }

    public void operator(String op) throws ParseException {
	if (isTextState()) {
	    textOutput(op);
	    return;
	}
	String code = MathMLTransTable.transCharacter( op, isEntity());
	if (code == null) {
	    System.err.println("["+op+"] is not defined.");
	    code = op;
	}
	addString( TagUtil.makeTag(this,"mo",code) );
    }

    public void number(String num) throws ParseException {
	if (isTextState()) {
	    textOutput(num);
	    return;
	}
	addString( TagUtil.makeTag(this,"mn",num) );
    }
    
    public void environment(EnvironmentContent env) throws ParseException {
	if (isTextState()) {
	    textOutput(env.getContent());
	    return;
	}
	MathMLEnvHandler handler = 
	    MathMLEnvTable.getHandler( env );
	if (handler != null) {
	    handler.onEnvBegin(this);
	    incTab();
	    handler.onContentOut(this);
	    decTab();
	    handler.onEnvEnd(this);
	    return;
	}
    }

    public void commandBegin(CommandContent cmd) throws ParseException {
	if (isTextState()) {
	    textOutput(cmd.getContent());
	    return;
	}

	//known command
	MathMLCommandHandler handler = 
	    MathMLCommandTable.getHandler( cmd );
	if (handler != null) {
	    handler.onCommandBegin(this);
	    handlerStack.push(handler);
	    int argn = handler.getArgNumber();
	    incTab();
	    return;
	}
	
	//no argument command --vv--
	handlerStack.push(null);

	String name = cmd.getContent();

	//operator command
	String op = MathMLTransTable.transCharacter( name, isEntity() );
	if (op != null) {
	    addString( TagUtil.makeTag(this,"mo",op) );
	    return;
	}
	
	//greek command
	String gr = MathMLTransTable.transGreek( name, isEntity() );
	if (gr != null) {
	    identifier_gen(gr);
	    return;
	}
	
	//function command
	String fn = MathMLTransTable.transFunction( name, isEntity() );
	if (fn != null) {
	    identifier_gen(fn);
	    //Amaya's bug?
	    //addString( makeTag("mo",transEntity("&ApplyFunction;"),curTab) );
	    return;
	}

	//unknown command
	// 1 : try to find from MathML Entity
	String ref = "&"+name+";";
	String ent = MathMLTransTable.transEntity( ref );
	if (ent != null && !ent.equals(ref)) {
	    log("unknown character ["+name+
			       "] : found in MathMLEntities.");
	    identifier_gen(TagUtil.transEntity(this,ref));
	    return;
	}
	// 2 : treat as a text
	addString( TagUtil.makeTag(this,"mtext",name));
	log("unknown command ["+name+
	    "] : not found MathMLEntities and TeX2MathML tables.");
	return;
    }

    public void commandOption() throws ParseException {
	MathMLCommandHandler handler = 
	    (MathMLCommandHandler)handlerStack.peek();

	if (handler == null || !handler.hasOption()) {
	    throw new InternalError(
		"["+handler.getClass().getName()+
		"] gone wrong way, at commandOption in MathMLVisitor.");
	}

	handler.onOption(this);
    }


    public void expressionBegin() {
	if (!isTextState()){
	    addString( TagUtil.makeBtag(this,"mrow") );
	    endStack.push( TagUtil.makeEtag(this,"mrow") );
	    incTab();
	}
    }

    public void expressionSubsupBegin() {
	if (!isTextState()){
	    addString( TagUtil.makeBtag(this,"msubsup") );
	    endStack.push( TagUtil.makeEtag(this,"msubsup") );
	    incTab();
	}
    }

    public void expressionSubBegin() {
	if (!isTextState()){
	    addString( TagUtil.makeBtag(this,"msub") );
	    endStack.push( TagUtil.makeEtag(this,"msub") );
	    incTab();
	}
    }

    public void expressionSupBegin() {
	if (!isTextState()){
	    addString( TagUtil.makeBtag(this,"msup") );
	    endStack.push( TagUtil.makeEtag(this,"msup") );
	    incTab();
	}
    }

    public void expressionEnd() {
	if (!isTextState()){
	    flushInterruptTag();
	    decTab();
	    String end = (String) endStack.pop();
	    addString( end );
	}
    }

    public void argBegin() {
	MathMLCommandHandler h = (MathMLCommandHandler) handlerStack.peek();
	if (h != null) 
	    h.onArgumentBegin(this);
	incTab();
    }

    public void argEnd() {
	decTab();
	MathMLCommandHandler h = (MathMLCommandHandler) handlerStack.peek();
	if (h != null) {
	    flushInterruptTag();
	    h.onArgumentEnd(this);
	}
    }

    public void commandEnd() {
	MathMLCommandHandler h = (MathMLCommandHandler) handlerStack.pop();
	if (h != null) {
	    flushInterruptTag();
	    decTab();
	    h.onCommandEnd(this);
	}
    }

    //==(util)====

    public void registInterruptTag(String etag) {
	interruptStack.push(etag);
	incTab();
    }
    
    protected void flushInterruptTag() {
	if (interruptStack.isEmpty()) return;
	while(!interruptStack.isEmpty()) {
	    String etag = (String)interruptStack.pop();
	    addString( etag );
	    decTab();
	}
    }

    public void setTextState(boolean b) {textState = b;}
    public boolean isTextState() {return textState ;}

    public void setStartTab(int i) {startTab = i;}
    public int getStartTab() {return startTab;}
    public int getCurrentTab() {return curTab;}

    public void incTab() {curTab++;}
    public void decTab() {curTab--;}

    public void log(String e) {
	log += e +"\n";
    }

    public void textOutput(String text) throws ParseException {
	int ps = 0;
	ps = text.indexOf("$");
	if (ps == -1) {
	    addString(TagUtil.tab(this)+text+TagUtil.getLS(this));
	    return;
	}
	//
	TeX2MathML eng = new TeX2MathML();
	MathMLCommandHandler curHandler = 
	    (MathMLCommandHandler)handlerStack.peek();
	while(true) {
	    addString(TagUtil.tab(this)+text.substring(0,ps)+
		      TagUtil.getLS(this));
	    text = text.substring(ps+1);
	    curHandler.onCommandEnd(this);
	    //enter math state
	    ps = text.indexOf("$");
	    if (ps == -1) {
		throw new InternalError("Can not find another <$>.");
	    }
	    //dirty code!
	    String math = text.substring(0,ps);
	    math = StringUtil.replace(math," ","");
	    math = TeXUtil.prepareTeXSource(math);
	    text = text.substring(ps+1);
	    addString(eng.translate_gen(math,this));
	    //end
	    curHandler.onCommandBegin(this);
	    ps = text.indexOf("$");
	    if (ps == -1) {
		addString(TagUtil.tab(this)+text+TagUtil.getLS(this));
		return;
	    }
	}
	
    }

    public void addString(String m) {
	buffer.append(m);
    }

}
