/* LaTeX2Sdoc
 *  Copyright (C) 2000-2001 SAKURAI, Masashi (m.sakurai@cmt.phys.kyushu-u.ac.jp)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package jp.gr.java_conf.ccs2.tool.tex2sdoc.math;
import jp.gr.java_conf.ccs2.tool.tex2sdoc.MacroManager;
import jp.gr.java_conf.ccs2.tool.tex2sdoc.TeXUtil;

/**
   translate math expression in LaTeX format 
   into MathML displaystyle format.
   (not thread safe)
 */
public class TeX2MathML {

    public TeX2MathML() {}

    /**
       translate math expression of LaTeX format
       into MathML displaystyle format.<br>
       
       For example,
       INPUT  : x^2+5x+6
       OUTPUT :
       &gt;mrow&lt;
         &gt;msup&lt;
	   &gt;mi&lt;x&gt;/mi&lt;
	   &gt;mn&lt;2&gt;/mn&lt;
	 &gt;/msup&lt;
         &gt;mo&lt;+&gt;/mo&lt;
         &gt;mn&lt;5&gt;/mn&lt;
         &gt;mi&lt;x&gt;/mi&lt;On the console, type 

         &gt;mo&lt;+&gt;/mo&lt;
         &gt;mn&lt;6&gt;/mn&lt;
       &gt;/mrow&lt;

       @param tex LaTeX math expression.
       @return MathML expression
     */
    public String translate(String tex) throws TeX2MathMLException {
	//normalize with separation of identifiers
	tex = TeXUtil.prepareTeXSource(tex);
	//expand macro
	if (getMacroManager() != null) {
	    tex = getMacroManager().expandMacro(tex);
	}
	//setting visitor properties
	MathMLVisitor v = getVisitor();
	v.setOneline( isOneline() );
	v.setEntity( isEntity() );
	v.setNameSpace( getNameSpace() );
	v.setMathMLURI( getMathMLURI() );
	//translate object
	try {
	    getCenter().traverse(getObjectTree(tex),v);
	} catch (ParseException e) {
	    throw new TeX2MathMLException(
		tex,e.getMessage(),getErrorRow(e),getErrorCol(e));
	} catch (TokenMgrError e) {
	    throw new TeX2MathMLException(
		tex,e.getMessage(),getErrorRow(e),getErrorCol(e));
	} catch (InternalError e) {
	    throw new TeX2MathMLException(
		tex,e.getMessage(),-1,-1);
	}
	log = v.getLog();
	return v.getCode();
    }

    static int getErrorRow(ParseException p) {
	return p.currentToken.next.beginLine;
    }
    static int getErrorCol(ParseException p) {
	return p.currentToken.next.beginColumn;
    }
    static int getErrorRow(TokenMgrError p) {
	return p.getErrorRows();
    }
    static int getErrorCol(TokenMgrError p) {
	return p.getErrorCols();
    }

    /** (for insider method)
       translate math expression of LaTeX format 
       into MathML displaystyle format.<br>
       The return value doesn't have <math> tag.
       And some initial prepareings are skiped.

       @see jp.gr.java_conf.ccs2.tool.tex2sdoc.math.handler.ArrayEnvHandler
       @param tex LaTeX math expression.
       @param tab tab number
       @return MathML expression
     */
    public String translate_gen(String tex,MathMLVisitor visitor) throws ParseException {
	//setting visitor properties
	MathMLVisitor v =  new MathMLVisitor(visitor);;
	//translate object
	tex = TeXUtil.prepareTeXSource(tex);
	getCenter().traverse(getObjectTree(tex),v);
	log = v.getLog();
	return v.getInsiderCode();
    }

    protected boolean oneline = false;
    /** 
	@param b if true, make a result without CR code.
	(defualt : false)
     */
    public void setOneline(boolean b) {oneline = b; }
    public boolean isOneline() { return oneline; }

    protected boolean entity = false;
    /** 
	@param b if true, make a result with Entity reference. 
	(default : false)
     */
    public void setEntity(boolean b) {entity = b;}
    public boolean isEntity() {return entity; }

    protected String nameSpace = null;
    /**
       if you want this translater to use name space,
       such as "&gt;mathml:mrow&lt;", set a cirtain name. (Ex. "mathml:")
       @param n prefix of MathML tag
       (default : null)
     */
    public void setNameSpace(String n) {
	nameSpace = n;
    }
    public String getNameSpace() {return nameSpace;}

    protected String mathmlUri = "http://www.w3.org/1998/Math/MathML";
    /** 
	set MathML URI. Default is "http://www.w3.org/1998/Math/MathML".
     */
    public void setMathMLURI(String s) {
	mathmlUri = s;
    }
    public String getMathMLURI() {return mathmlUri;}

    protected MacroManager macros;
    /**
       set MacroManager. Default is null.
       If you set a MacroManager object, this will
       expand macros in the given TeX source.
     */
    public void setMacroManager(MacroManager m) {
	macros = m;
    }
    public MacroManager getMacroManager() {return macros;}

    protected String log = "";
    /** get translation error log */
    public String getLog() {return log;}

    //===(frame work area)=============

    public TeXSourceVisitorCenter getCenter() {
	if (center == null)
	    center = new TeXSourceVisitorCenter();
	return center;
    }
    public MathMLVisitor getVisitor() {
	if (visitor == null)
	    visitor = new MathMLVisitor();
	return visitor;
    }

    public Expression getObjectTree(String tex) throws ParseException {
	TeXFormula f = new TeXFormula(tex);
	return f.getExpression();
    }

    protected MathMLVisitor visitor;
    protected TeXSourceVisitorCenter center;

}
